#!/usr/bin/env python3
# THIS FILE IS CURSED, DO NOT LOOK AT IT!!!

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=import-outside-toplevel

"""Nobody inspects the spammish repetition."""

from __future__ import annotations

import os
import sys
import tarfile
import typing
import zipfile
from collections.abc import Callable, Mapping, Set
from datetime import datetime, timedelta, timezone
from functools import cache
from importlib.metadata import Distribution
from importlib.util import module_from_spec, spec_from_file_location
from os import PathLike
from pathlib import Path
from warnings import filterwarnings

from setuptools import setup

BACKEND_REQUIRES: set[str] = set()
DULWICH = "dulwich"
GET_VERSION = "get-version"
TROVE_CLASSIFIERS = "trove-classifiers"
TIME_MACHINE = "time-machine"
ZOPFLIPY = "zopflipy"

WHEEL_BUILD_DEPS: Set[str] = {TIME_MACHINE}

filterwarnings("ignore", "", UserWarning, "setuptools.dist")

HELP = "--help" in sys.argv[1:]
VERSION = "--version" in sys.argv[1:]
BUILDING = not HELP and {"sdist", "bdist_wheel"} & {*sys.argv[1:]}

classifiers = [
    "Development Status :: 5 - Production/Stable",
    (
        "License :: OSI Approved :: "
        "GNU Affero General Public License v3 or later (AGPLv3+)"
    ),
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: Implementation :: CPython",
    "Typing :: Typed",
]

os.environ.update(
    GIT_CONFIG_COUNT="1",
    GIT_CONFIG_KEY_0="core.abbrev",
    GIT_CONFIG_VALUE_0="10",
)


@cache
def get_constraints() -> Mapping[str, str]:
    """Get the constraints for the libraries."""
    constraints_file = path("pip-constraints.txt")
    if not constraints_file.exists():
        constraints_file = path("CNSTRNTS.TXT")

    return {
        line.split("==")[0].strip(): line.strip()
        for line in constraints_file.read_text("UTF-8").splitlines()
    }


def get_version() -> str:
    """Get the version."""
    if path(".git").exists():
        BACKEND_REQUIRES.add(GET_VERSION)
        try:
            # pylint: disable=redefined-outer-name
            from get_version import get_version

            version = get_version(__file__, vcs="git")
            path("VERSIONS.TXT").write_text(version + "\n")
        except ImportError:
            return "MissingNo."
    if path("VERSIONS.TXT").is_file():
        return path("VERSIONS.TXT").read_text("UTF-8").strip()
    return Distribution.at(path(".")).version


def path(path: str | PathLike[str]) -> Path:
    """Return the absolute path to the file."""
    # pylint: disable=redefined-outer-name
    return Path(__file__).resolve().parent / path


if path("pip-constraints.txt").exists() and (BUILDING or VERSION):
    path("CNSTRNTS.TXT").write_text(
        "\n".join(sorted(get_constraints()[dep] for dep in WHEEL_BUILD_DEPS))
        + "\n",
        encoding="UTF-8",
    )

    path("TESTDEPS.TXT").write_text(
        "\n".join(
            sorted(
                line
                for line in path("pip-dev-requirements.txt")
                .read_text("UTF-8")
                .splitlines()
                if line.startswith(
                    ("pytest", "html5lib==", "time-machine==", "zstandard==")
                )
            )
        )
        + "\n",
        encoding="UTF-8",
    )


if not path(".git").exists():
    pass
elif not (BUILDING or VERSION):
    BACKEND_REQUIRES.add(DULWICH)
    BACKEND_REQUIRES.add(TROVE_CLASSIFIERS)
else:
    from dulwich.repo import Repo

    repo = Repo(path(".").as_posix())
    path("REVISION.TXT").write_bytes(repo.head())
    head = repo[repo.head()]
    dt = datetime.fromtimestamp(
        head.commit_time, timezone(timedelta(seconds=head.commit_timezone))
    )
    path("TIMESTMP.TXT").write_text(dt.isoformat(), encoding="UTF-8")
    del dt, head, Repo
    with path("an_website/static/commits.txt").open("wb") as file:
        for entry in repo.get_walker():
            file.write(entry.commit.id)
            file.write(b" ")
            file.write(str(entry.commit.commit_time).encode("UTF-8"))
            file.write(b" ")
            file.write(entry.commit.message.split(b"\n")[0])
            file.write(b"\n")
            del entry
        file.flush()
    del repo, file

    if not VERSION:
        import trove_classifiers as trove

        assert all(_ in trove.classifiers for _ in classifiers)
        assert classifiers == sorted(classifiers)

        from zopfli import ZipFile

        zipfile.ZipFile = ZipFile  # type: ignore[assignment, misc]


BACKEND_REQUIRES.add(TIME_MACHINE)
if BUILDING:
    import time_machine

    time_machine.travel(
        path("TIMESTMP.TXT").read_text("UTF-8"), tick=False
    ).start()

    os.environ["SOURCE_DATE_EPOCH"] = str(int(datetime.now().timestamp()))

    class Tarfile(tarfile.TarFile):
        """Tarfile sub-class."""

        def add(self, *args: typing.Any, **kwargs: typing.Any) -> None:
            """Add stuff."""
            orig_filter: Callable[[tarfile.TarInfo], tarfile.TarInfo] = (
                kwargs.get("filter", lambda _: _)
            )

            def filter_(tarinfo: tarfile.TarInfo) -> tarfile.TarInfo:
                tarinfo = orig_filter(tarinfo)
                tarinfo.mtime = datetime.now().timestamp()
                tarinfo.gid = 0
                tarinfo.gname = ""
                tarinfo.uid = 0
                tarinfo.uname = ""
                return tarinfo

            kwargs["filter"] = filter_
            return super().add(*args, **kwargs)

    tarfile.open = Tarfile.open


compress_script_path = path("scripts/compress_static_files.py")
if compress_script_path.exists():
    BACKEND_REQUIRES.add(ZOPFLIPY)
    compress_spec = spec_from_file_location(
        "compress_static_files", compress_script_path
    )
    assert compress_spec and compress_spec.loader
    compress_module = module_from_spec(compress_spec)
    compress_spec.loader.exec_module(compress_module)
    BACKEND_REQUIRES.update(compress_module.get_missing_dependencies())
    if BUILDING:
        for _ in compress_module.compress_static_files():
            pass
    del compress_spec, compress_module
del compress_script_path

install_requires = path("pip-requirements.txt").read_text("UTF-8").split("\n")
long_description = path("README.md").read_text("UTF-8")

dist = setup(
    author="Das Asoziale Netzwerk",
    author_email="contact@asozial.org",
    classifiers=classifiers,
    description="#1 Website in the Worlds",
    include_package_data=True,
    install_requires=install_requires,
    license="AGPL-3.0-or-later",
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="an-website",
    packages=["an_website"],
    platforms="OS Independent",
    python_requires=">=3.12",
    url="https://github.com/asozialesnetzwerk/an-website",
    version=get_version(),
    zip_safe=True,
    setup_requires=[get_constraints()[dep] for dep in BACKEND_REQUIRES],
    entry_points={
        "console_scripts": (
            "an-website = an_website.__main__:main",
            "an-backdoor-client = an_website.backdoor_client:main",
        )
    },
)

for t, _, dist_file in dist.dist_files:
    if t != "sdist":
        continue
    if not dist_file.endswith(".gz"):
        continue

    import zopfli

    f = zopfli.ZOPFLI_FORMAT_GZIP
    d = zopfli.ZopfliDecompressor(f)
    data = d.decompress(Path(dist_file).read_bytes()) + d.flush()
    c = zopfli.ZopfliCompressor(f)
    Path(dist_file).write_bytes(c.compress(data) + c.flush())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page with the kangaroo comics by Zeit Online."""

from __future__ import annotations

from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/kaenguru-comics", KangarooComicsRedirect),
            (r"/kaenguru-comics-alt", KangarooComics),
        ),
        name="Känguru-Comics",
        description=(
            "Känguru-Comics von Zeit Online, Marc-Uwe Kling und Bernd Kissel"
        ),
        path="/kaenguru-comics",
        keywords=("Känguru", "Comics", "Zeit", "Marc-Uwe Kling"),
        aliases=(
            "/kangaroo-comics",
            "/comics",
            "/känguru-comics",
            "/k%C3%A4nguru-comics",
        ),
    )


class KangarooComicsRedirect(HTMLRequestHandler):
    """Request handler for the kangaroo comics redirect."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the kangaroo comics page."""
        if head:
            return
        await self.render(
            "pages/redirect.html",
            send_referrer=False,
            redirect_url="https://www.zeit.de/serie/die-kaenguru-comics",
            from_url=self.fix_url("/kaenguru-comics-alt"),
            back_button_text="Comics hier lesen (nicht empfohlen)",
            discord=False,
        )


class KangarooComics(HTMLRequestHandler):
    """Request handler for the kangaroo comics page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the kangaroo comics page."""
        if head:
            return
        await self.render("pages/kangaroo_comics.html")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the quotes pages."""

from __future__ import annotations

import urllib.parse
from datetime import datetime, timezone as tz
from functools import cache
from io import BytesIO

import orjson as json
import qoi_rs
from PIL import Image

from an_website.quotes import create, utils as quotes
from an_website.quotes.image import CONTENT_TYPES, FILE_EXTENSIONS

from . import (  # noqa: F401  # pylint: disable=unused-import
    WRONG_QUOTE_DATA,
    FetchCallable,
    app,
    assert_valid_html_response,
    assert_valid_json_response,
    assert_valid_redirect,
    assert_valid_response,
    check_html_page,
    fetch,
)


@cache
def get_wrong_quote() -> quotes.WrongQuote:
    """Patch stuff."""
    return quotes.parse_wrong_quote(WRONG_QUOTE_DATA)


async def test_parsing_wrong_quotes() -> None:
    """Test parsing wrong_quotes."""
    wrong_quote = get_wrong_quote()

    assert wrong_quote.id == 1
    # quote_id (1) - author_id (2)
    assert wrong_quote.get_id_as_str() == "1-2"
    assert wrong_quote.rating == 4

    # parsing the same dict twice should return the same object twice
    assert id(wrong_quote) == id(quotes.parse_wrong_quote(WRONG_QUOTE_DATA))

    author = quotes.parse_author(WRONG_QUOTE_DATA["author"])  # type: ignore[arg-type]
    assert id(author) == id(wrong_quote.author)
    assert author == await quotes.get_author_by_id(author_id=author.id)
    assert author.name == "Kim Jong-il"
    assert author.id == 2

    quote = quotes.parse_quote(WRONG_QUOTE_DATA["quote"])  # type: ignore[arg-type]
    assert id(quote) == id(wrong_quote.quote)
    assert quote == await quotes.get_quote_by_id(quote_id=quote.id)
    assert quote.id == 1
    assert quote.author.id == 1

    assert await quotes.get_rating_by_id(1, 2) == 4

    assert len(quotes.QUOTES_CACHE) == 1
    assert quotes.MAX_QUOTES_ID.value == 1
    assert len(quotes.AUTHORS_CACHE) == 2
    assert quotes.MAX_AUTHORS_ID.value == 2


def test_author_updating() -> None:
    """Test updating the author."""
    get_wrong_quote()

    assert (
        author := quotes.parse_author({"id": 1, "author": "test"})
    ).name == "test"

    quotes.parse_wrong_quote(WRONG_QUOTE_DATA)

    assert author.name == "Abraham Lincoln"


async def test_create() -> None:
    """Test some functions in the quote create module."""
    wrong_quote = get_wrong_quote()

    assert wrong_quote.quote.author is create.get_author_by_name(
        "Abraham Lincoln"
    )
    assert [wrong_quote.quote.author] == await create.get_authors(
        "Abraham Lincoln"
    )
    assert wrong_quote.quote.author is await create.create_author(
        "Abraham Lincoln"
    )
    assert wrong_quote.quote.author in await create.get_authors("abrah lincoln")

    assert wrong_quote.author is create.get_author_by_name("Kim Jong-il")
    assert wrong_quote.author is create.get_author_by_name("kIm jong-il")
    assert wrong_quote.author is await create.create_author("kIm jong-il")

    assert [wrong_quote.author] == await create.get_authors("kIm jong-il")
    assert wrong_quote.author in await create.get_authors("kIn jon il")

    assert create.get_author_by_name("lol") is None

    quote_str = wrong_quote.quote.quote.lower()

    assert create.get_quote_by_str(quote_str) is wrong_quote.quote
    assert create.get_quote_by_str(quote_str.upper()) is wrong_quote.quote
    assert create.get_quote_by_str(quote_str.title()) is wrong_quote.quote
    assert create.get_quote_by_str(f"„{quote_str}“") is wrong_quote.quote

    assert (
        await create.create_quote(
            f'"{quote_str}"',  # noqa: B907
            None,  # type: ignore[arg-type]
        )
        is wrong_quote.quote
    )

    assert await create.get_quotes(quote_str) == [wrong_quote.quote]

    modified_quote_str = quote_str[1:8] + "x" + quote_str[9:-1].upper()
    assert wrong_quote.quote in await create.get_quotes(modified_quote_str)


async def test_argument_checking_create_pages(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test whether the create handlers complain because of missing args."""
    wrong_quote = get_wrong_quote()

    await quotes.make_api_request("/test", entity_should_exist=False)

    for data in (
        "quote-1=&fake-author-1=",
        "quote-1=test&fake-author-1=",
        "quote-1=&fake-author-1=test",
        "x=y",
    ):
        assert_valid_html_response(
            await fetch("/zitate/erstellen", method="POST", body=data), {400}
        )

    for data in (
        "quote-2=&fake-author-2=",
        "quote-2=test&fake-author-2=",
        "quote-2=&fake-author-2=test",
        "x=y",
    ):
        assert_valid_html_response(
            await fetch("/zitate/create-wrong-quote", method="POST", body=data),
            {400},
        )

    for num in (1, 2):
        url = "/zitate/erstellen" if num == 1 else "/zitate/create-wrong-quote"
        await assert_valid_redirect(
            fetch,
            url,
            "/zitate/1-1",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=json.dumps(
                {
                    f"quote-{num}": wrong_quote.quote.quote.lower(),
                    f"fake-author-{num}": wrong_quote.quote.author.name.upper(),
                }
            ),
            codes={303},
        )
        await assert_valid_redirect(
            fetch,
            url,
            "/zitate/1-2",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=json.dumps(
                {
                    f"quote-{num}": wrong_quote.quote.quote.upper(),
                    f"fake-author-{num}": wrong_quote.author.name.lower(),
                }
            ),
            codes={303},
        )


async def test_quote_updating() -> None:
    """Test updating the quote."""
    get_wrong_quote()

    quote = await quotes.get_quote_by_id(1)

    assert (
        quote.quote == "Frage nicht, was dein Land für dich tun kann, "
        "frage, was du für dein Land tun kannst."
    )
    quote.quote = "test"

    assert quote.quote == "test"

    quotes.parse_wrong_quote(WRONG_QUOTE_DATA)

    assert (
        quote.quote == "Frage nicht, was dein Land für dich tun kann, "
        "frage, was du für dein Land tun kannst."
    )


async def test_quote_request_handlers(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test the request handlers for the quotes page."""
    get_wrong_quote()  # add data to cache
    await check_html_page(fetch, "/zitate")
    await check_html_page(fetch, "/zitate/1-1")
    await check_html_page(fetch, "/zitate/1-2")

    assert_valid_json_response(await fetch("/api/zitate/1-1"))
    assert_valid_json_response(await fetch("/api/zitate/1-2"))

    await assert_valid_redirect(
        fetch, "/zitate?quote=1&author=1", "/zitate/1-1"
    )
    await assert_valid_redirect(fetch, "/z/1-1", "/zitate/1-1")

    await assert_valid_redirect(fetch, "/z/1", "/zitate/1")
    await assert_valid_redirect(fetch, "/zitate/1", "/zitate/1-2")

    await assert_valid_redirect(fetch, "/zitate?author=1", "/zitate/-1")
    await assert_valid_redirect(fetch, "/z/-1", "/zitate/-1")
    await assert_valid_redirect(fetch, "/zitate/-1", "/zitate/info/a/1")

    await assert_valid_redirect(fetch, "/zitate?quote=1", "/zitate/1-")
    await assert_valid_redirect(fetch, "/z/1-", "/zitate/1-")
    await assert_valid_redirect(fetch, "/zitate/1-", "/zitate/info/z/1")

    for i in (1, 2):
        # twice because we cache the author info from wikipedia
        assert_valid_html_response(await fetch(f"/zitate/info/a/{i}"))
        assert_valid_html_response(await fetch(f"/zitate/info/a/{i}"))

    await check_html_page(fetch, "/zitate/info/z/1")
    await check_html_page(fetch, "/zitate/share/1-1")
    await check_html_page(fetch, "/zitate/erstellen")
    assert not assert_valid_response(
        await fetch("/zitate/info/z/1", method="HEAD"),
        "text/html;charset=utf-8",
    ).body
    assert not assert_valid_response(
        await fetch("/zitate/share/1-1", method="HEAD"),
        "text/html;charset=utf-8",
    ).body
    assert not assert_valid_response(
        await fetch("/zitate/erstellen", method="HEAD"),
        "text/html;charset=utf-8",
    ).body

    assert len(
        assert_valid_response(await fetch("/zitate/1-1.gif"), "image/gif").body
    ) > len(
        assert_valid_response(
            await fetch("/zitate/1-1.gif?small=sure"), "image/gif"
        ).body
    )


async def test_quote_image_handlers(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Text downloading quotes as images."""
    for ext, name in FILE_EXTENSIONS.items():
        content_type = CONTENT_TYPES[name]
        image1 = assert_valid_response(
            await fetch(f"/zitate/1-1.{ext}"), content_type
        ).body
        image2 = assert_valid_response(
            await fetch("/zitate/1-1", headers={"Accept": content_type}),
            content_type,
        ).body
        image3 = assert_valid_response(
            await fetch("/api/zitate/1-1", headers={"Accept": content_type}),
            content_type,
        ).body
        if name == "txt":
            assert image1 == image3
        else:
            image4 = assert_valid_response(
                await fetch(
                    "/zitate/1-1/image", headers={"Accept": content_type}
                ),
                content_type,
            ).body
            if name == "xlsx":
                continue
            assert image1 == image2 == image3 == image4
            if name == "pdf":
                continue
            img = Image.open(BytesIO(image4), formats=[name])
            try:  # pylint: disable=too-many-try-statements
                img.load()
                assert img.width == 1000
                assert img.height == 750
                if name == "gif":
                    assert img.getpixel((0, 0)) == 255
                elif name == "spider":
                    assert not img.getpixel((0, 0))
                else:
                    assert img.getpixel((0, 0)) == (0, 0, 0), f"{name}"
                if ext == "qoi":
                    img2 = qoi_rs.decode_pillow(image4)
                    assert tuple(img.getdata()) == tuple(img2.getdata())
                    img2.close()
                    qimg = qoi_rs.decode(image4)
                    assert qimg.mode == "RGB"
                    assert qimg.width == 1000
                    assert qimg.height == 750
            finally:
                img.close()


async def test_quote_redirect_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the quote redirect API."""
    response = await fetch(
        "/api/zitate?show-rating=s&r=all", follow_redirects=True
    )
    url = urllib.parse.urlsplit(response.effective_url)
    assert "show-rating=sure" in url.query
    assert "r=all" in url.query
    json_ = assert_valid_json_response(response)
    assert url.path == f"/api/zitate/{json_['id']}"
    assert json_["rating"] != "???"


async def test_quote_apis(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the quote APIs."""
    wrong_quote = get_wrong_quote()
    today = datetime.now(tz=tz.utc).date()

    response = assert_valid_json_response(
        await fetch(f"/api/zitate/{wrong_quote.get_id_as_str()}?show-rating=s")
    )
    if (today.month, today.day) != (4, 1):
        assert response["id"] == wrong_quote.get_id_as_str()
        assert response["quote"] == wrong_quote.quote.quote
        assert response["author"] == wrong_quote.author.name
        assert response["real_author"] == wrong_quote.quote.author.name
        assert response["real_author_id"] == wrong_quote.quote.author.id
        assert int(response["rating"]) == wrong_quote.rating

    response = assert_valid_json_response(
        await fetch(
            f"/api/zitate/{wrong_quote.get_id_as_str()}/full?show-rating=sure"
        )
    )
    if (today.month, today.day) != (4, 1):
        assert response["wrong_quote"] == wrong_quote.to_json()
        response = response["wrong_quote"]
        assert response["id"] == wrong_quote.get_id_as_str()
        assert response["quote"] == wrong_quote.quote.to_json()
        assert response["author"] == wrong_quote.author.to_json()
        assert response["path"] == f"/zitate/{wrong_quote.get_id_as_str()}"
        assert int(response["rating"]) == wrong_quote.rating

    for count in (6, 2, 1):
        response = assert_valid_json_response(
            await fetch(f"/api/zitate/generator?count={count}")
        )
        assert 0 < len(response["quotes"]) <= count
        assert 0 < len(response["authors"]) <= count

        for quote in response["quotes"]:
            assert quote == quotes.QUOTES_CACHE[quote["id"]].to_json()

        for author in response["authors"]:
            assert author == quotes.AUTHORS_CACHE[author["id"]].to_json()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The main page of the website."""

from __future__ import annotations

import contextlib

from ..quotes.quote_of_the_day import QuoteOfTheDayBaseHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        # the empty dict prevents the header from being changed
        handlers=((r"/", MainPage, {}),),
        name="Hauptseite",
        description="Die Hauptseite der Webseite",
        path="/",
        keywords=("asozial", "Asoziales", "Netzwerk"),
    )


class MainPage(QuoteOfTheDayBaseHandler):
    """The request handler of the main page."""

    async def check_ready(self) -> None:  # noqa: D102
        pass

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the main page."""
        if head:
            return
        quote_data = None
        if self.redis:  # type: ignore[truthy-bool]
            with contextlib.suppress(Exception):
                quote_data = await self.get_quote_of_today()
        await self.render("pages/main_page.html", quote_data=quote_data)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A 🆒 chat."""

from __future__ import annotations

import asyncio
import logging
import random
import sys
import time
from collections.abc import Awaitable, Iterable, Mapping
from typing import Any, Final, Literal

import orjson as json
from emoji import EMOJI_DATA, demojize, emoji_list, emojize, purely_emoji
from redis.asyncio import Redis
from tornado.web import Application, HTTPError
from tornado.websocket import WebSocketHandler

from .. import EPOCH_MS, EVENT_REDIS, EVENT_SHUTDOWN, NAME, ORJSON_OPTIONS
from ..utils.base_request_handler import BaseRequestHandler
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import Permission, ratelimit
from .pub_sub_provider import PubSubProvider

LOGGER: Final = logging.getLogger(__name__)

EMOJIS_NO_FLAGS: Final[tuple[str, ...]] = tuple(
    emoji
    for emoji in EMOJI_DATA
    if ord(emoji[0]) not in range(0x1F1E6, 0x1F200)
)

MAX_MESSAGE_SAVE_COUNT: Final = 100
MAX_MESSAGE_LENGTH: Final = 20
REDIS_CHANNEL: Final = f"{NAME}:emoji_chat_channel"


def get_ms_timestamp() -> int:
    """Get the current time in ms."""
    return time.time_ns() // 1_000_000 - EPOCH_MS


async def subscribe_to_redis_channel(
    app: Application, worker: int | None
) -> None:
    """Subscribe to the Redis channel and handle incoming messages."""
    get_pubsub = PubSubProvider((REDIS_CHANNEL,), app.settings, worker)
    del app

    while not EVENT_SHUTDOWN.is_set():  # pylint: disable=while-used
        ps = await get_pubsub()
        try:
            message = await ps.get_message(timeout=5.0)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            if str(exc) == "Connection closed by server.":
                continue
            LOGGER.exception("Failed to get message on worker %s", worker)
            await asyncio.sleep(0)
            continue

        match message:
            case None:
                pass
            case {
                "type": "message",
                "data": str(data),
                "channel": channel,
            } if channel == REDIS_CHANNEL:
                await asyncio.gather(
                    *[conn.write_message(data) for conn in OPEN_CONNECTIONS]
                )
            case {
                "type": "subscribe",
                "data": 1,
                "channel": channel,
            } if channel == REDIS_CHANNEL:
                logging.info(
                    "Subscribed to Redis channel %r on worker %s",
                    channel,
                    worker,
                )
            case _:
                logging.error(
                    "Got unexpected message %s on worker %s",
                    message,
                    worker,
                )
        await asyncio.sleep(0)


async def save_new_message(
    author: str,
    message: str,
    redis: Redis[str],
    redis_prefix: str,
) -> None:
    """Save a new message."""
    message_dict = {
        "author": [data["emoji"] for data in emoji_list(author)],
        "content": [data["emoji"] for data in emoji_list(message)],
        "timestamp": get_ms_timestamp(),
    }
    await redis.rpush(
        f"{redis_prefix}:emoji-chat:message-list",
        json.dumps(message_dict, option=ORJSON_OPTIONS),
    )
    await redis.ltrim(
        f"{redis_prefix}:emoji-chat:message-list", -MAX_MESSAGE_SAVE_COUNT, -1
    )
    LOGGER.info("GOT new message %s", message_dict)
    await redis.publish(
        REDIS_CHANNEL,
        json.dumps(
            {
                "type": "message",
                "message": message_dict,
            },
            option=ORJSON_OPTIONS,
        ),
    )


async def get_messages(
    redis: Redis[str],
    redis_prefix: str,
    start: None | int = None,
    stop: int = -1,
) -> list[dict[str, Any]]:
    """Get the messages."""
    start = start if start is not None else -MAX_MESSAGE_SAVE_COUNT
    messages = await redis.lrange(
        f"{redis_prefix}:emoji-chat:message-list", start, stop
    )
    return [json.loads(message) for message in messages]


def check_message_invalid(message: str) -> Literal[False] | str:
    """Check if a message is an invalid message."""
    if not message:
        return "Empty message not allowed."

    if not purely_emoji(message):
        return "Message can only contain emojis."

    if len(emoji_list(message)) > MAX_MESSAGE_LENGTH:
        return f"Message longer than {MAX_MESSAGE_LENGTH} emojis."

    return False


def emojize_user_input(string: str) -> str:
    """Emojize user input."""
    string = emojize(string, language="de")
    string = emojize(string, language="en")
    string = emojize(string, language="alias")
    return string


def normalize_emojis(string: str) -> str:
    """Normalize emojis in a string."""
    return emojize(demojize(string))


def get_random_name() -> str:
    """Generate a random name."""
    return normalize_emojis(
        "".join(random.sample(EMOJIS_NO_FLAGS, 5))  # nosec: B311
    )


class ChatHandler(BaseRequestHandler):
    """The request handler for the emoji chat."""

    RATELIMIT_GET_BUCKET = "emoji-chat-get-messages"
    RATELIMIT_GET_LIMIT = 10
    RATELIMIT_GET_COUNT_PER_PERIOD = 10
    RATELIMIT_GET_PERIOD = 1

    RATELIMIT_POST_BUCKET = "emoji-chat-send-message"
    RATELIMIT_POST_LIMIT = 5
    RATELIMIT_POST_COUNT_PER_PERIOD = 5
    RATELIMIT_POST_PERIOD = 5

    async def get(
        self,
        *,
        head: bool = False,
    ) -> None:
        """Show the users the current messages."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)

        if head:
            return

        await self.render_chat(
            await get_messages(self.redis, self.redis_prefix)
        )

    async def get_name(self) -> str:
        """Get the name of the user."""
        cookie = self.get_secure_cookie(
            "emoji-chat-name",
            max_age_days=90,
            min_version=2,
        )

        name = cookie.decode("UTF-8") if cookie else get_random_name()

        # save it in cookie or reset expiry date
        if not self.get_secure_cookie(
            "emoji-chat-name", max_age_days=30, min_version=2
        ):
            self.set_secure_cookie(
                "emoji-chat-name",
                name.encode("UTF-8"),
                expires_days=90,
                path="/",
                samesite="Strict",
            )

        geoip = await self.geoip() or {}
        if "country_flag" in geoip:
            flag = geoip["country_flag"]
        elif self.request.host_name.endswith(".onion"):
            flag = "🏴‍☠"
        else:
            flag = "❔"

        return normalize_emojis(name + flag)

    async def get_name_as_list(self) -> list[str]:
        """Return the name as list of emojis."""
        return [emoji["emoji"] for emoji in emoji_list(await self.get_name())]

    async def post(self) -> None:
        """Let users send messages and show the users the current messages."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)

        message = emojize_user_input(
            normalize_emojis(self.get_argument("message"))
        )

        if err := check_message_invalid(message):
            raise HTTPError(400, reason=err)

        await save_new_message(
            await self.get_name(),
            message,
            redis=self.redis,
            redis_prefix=self.redis_prefix,
        )

        await self.render_chat(
            await get_messages(self.redis, self.redis_prefix)
        )

    async def render_chat(self, messages: Iterable[Mapping[str, Any]]) -> None:
        """Render the chat."""
        raise NotImplementedError


class HTMLChatHandler(ChatHandler, HTMLRequestHandler):
    """The HTML request handler for the emoji chat."""

    async def render_chat(self, messages: Iterable[Mapping[str, Any]]) -> None:
        """Render the chat."""
        await self.render(
            "pages/emoji_chat.html",
            messages=messages,
            user_name=await self.get_name_as_list(),
        )


class APIChatHandler(ChatHandler, APIRequestHandler):
    """The API request handler for the emoji chat."""

    async def render_chat(self, messages: Iterable[Mapping[str, Any]]) -> None:
        """Render the chat."""
        await self.finish(
            {
                "current_user": await self.get_name_as_list(),
                "messages": messages,
            }
        )


OPEN_CONNECTIONS: list[ChatWebSocketHandler] = []


class ChatWebSocketHandler(WebSocketHandler, ChatHandler):
    """The handler for the chat WebSocket."""

    name: str
    connection_time: int

    def on_close(self) -> None:  # noqa: D102
        LOGGER.info("WebSocket closed")
        OPEN_CONNECTIONS.remove(self)
        for conn in OPEN_CONNECTIONS:
            conn.send_users()

    def on_message(self, message: str | bytes) -> Awaitable[None] | None:
        """Respond to an incoming message."""
        if not message:
            return None
        message2: dict[str, Any] = json.loads(message)
        if message2["type"] == "message":
            if "message" not in message2:
                return self.write_message(
                    {
                        "type": "error",
                        "error": "Message needs message key with the message.",
                    }
                )
            return self.save_new_message(message2["message"])

        return self.write_message(
            {"type": "error", "error": f"Unknown type {message2['type']}."}
        )

    async def open(self, *args: str, **kwargs: str) -> None:
        # pylint: disable=invalid-overridden-method
        """Handle an opened connection."""
        LOGGER.info("WebSocket opened")
        await self.write_message(
            {
                "type": "init",
                "current_user": [
                    emoji["emoji"] for emoji in emoji_list(self.name)
                ],
            }
        )

        self.connection_time = get_ms_timestamp()
        OPEN_CONNECTIONS.append(self)
        for conn in OPEN_CONNECTIONS:
            conn.send_users()

        await self.send_messages()

    async def prepare(self) -> None:  # noqa: D102
        self.now = await self.get_time()

        if not EVENT_REDIS.is_set():
            raise HTTPError(503)

        self.name = await self.get_name()

        if not await self.ratelimit(True):
            await self.ratelimit()

    async def render_chat(self, messages: Iterable[Mapping[str, Any]]) -> None:
        """Render the chat."""
        raise NotImplementedError

    async def save_new_message(self, msg_text: str) -> None:
        """Save a new message."""
        msg_text = emojize_user_input(normalize_emojis(msg_text).strip())
        if err := check_message_invalid(msg_text):
            return await self.write_message({"type": "error", "error": err})

        if self.settings.get("RATELIMITS") and not self.is_authorized(
            Permission.RATELIMITS
        ):
            if not EVENT_REDIS.is_set():
                return await self.write_message({"type": "ratelimit"})

            ratelimited, headers = await ratelimit(
                self.redis,
                self.redis_prefix,
                str(self.request.remote_ip),
                bucket=self.RATELIMIT_POST_BUCKET,
                max_burst=self.RATELIMIT_POST_LIMIT - 1,
                count_per_period=self.RATELIMIT_POST_COUNT_PER_PERIOD,
                period=self.RATELIMIT_POST_PERIOD,
                tokens=1,
            )

            if ratelimited:
                return await self.write_message(
                    {"type": "ratelimit", "retry_after": headers["Retry-After"]}
                )

        return await save_new_message(
            self.name, msg_text, self.redis, self.redis_prefix
        )

    async def send_messages(self) -> None:
        """Send this WebSocket all current messages."""
        return await self.write_message(
            {
                "type": "messages",
                "messages": await get_messages(self.redis, self.redis_prefix),
            },
        )

    def send_users(self) -> None:
        """Send this WebSocket all current users."""
        if sys.flags.dev_mode:
            self.write_message(  # type: ignore[unused-awaitable]
                {
                    "type": "users",
                    "users": [
                        {
                            "name": conn.name,
                            "joined_at": conn.connection_time,
                        }
                        for conn in OPEN_CONNECTIONS
                    ],
                }
            )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=preferred-module

"""Nobody inspects the spammish repetition."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import fields, is_dataclass
from datetime import date, datetime, time, timezone
from enum import Enum
from functools import partial
from json import JSONDecodeError, dumps as _dumps, loads as _loads
from typing import Any
from uuid import UUID

__version__ = "3.8.14"  # TODO: orjson.Fragment

__all__ = (
    "dumps",
    "loads",
    "JSONDecodeError",
    "JSONEncodeError",
    "OPT_APPEND_NEWLINE",
    "OPT_INDENT_2",
    "OPT_NAIVE_UTC",
    "OPT_NON_STR_KEYS",
    "OPT_OMIT_MICROSECONDS",
    "OPT_PASSTHROUGH_DATACLASS",
    "OPT_PASSTHROUGH_DATETIME",
    "OPT_PASSTHROUGH_SUBCLASS",
    "OPT_SERIALIZE_DATACLASS",
    "OPT_SERIALIZE_NUMPY",
    "OPT_SERIALIZE_UUID",
    "OPT_SORT_KEYS",
    "OPT_STRICT_INTEGER",
    "OPT_UTC_Z",
)

OPT_APPEND_NEWLINE = 1 << 10
OPT_INDENT_2 = 1 << 0
OPT_NAIVE_UTC = 1 << 1
OPT_NON_STR_KEYS = 1 << 2  # TODO: require for non-string keys
OPT_OMIT_MICROSECONDS = 1 << 3
OPT_PASSTHROUGH_DATACLASS = 1 << 11
OPT_PASSTHROUGH_DATETIME = 1 << 9
OPT_PASSTHROUGH_SUBCLASS = 1 << 8  # TODO
OPT_SERIALIZE_DATACLASS = 0
OPT_SERIALIZE_NUMPY = 1 << 4  # TODO
OPT_SERIALIZE_UUID = 0
OPT_SORT_KEYS = 1 << 5
OPT_STRICT_INTEGER = 1 << 6  # TODO
OPT_UTC_Z = 1 << 7


class JSONEncodeError(TypeError):
    """Nobody inspects the spammish repetition."""


def _default(  # pylint: disable=too-complex
    default: None | Callable[[Any], Any], option: int, spam: object
) -> Any:
    serialize_dataclass = not option & OPT_PASSTHROUGH_DATACLASS
    serialize_datetime = not option & OPT_PASSTHROUGH_DATETIME

    if serialize_dataclass and is_dataclass(spam):
        return {
            field.name: getattr(spam, field.name)
            for field in fields(spam)
            if not field.name.startswith("_")
        }
    if serialize_datetime and type(spam) in (date, datetime, time):
        if type(spam) in (datetime, time):
            if option & OPT_OMIT_MICROSECONDS:
                spam = spam.replace(microsecond=0)  # type: ignore[attr-defined]
            if type(spam) is datetime:  # pylint: disable=unidiomatic-typecheck
                if option & OPT_NAIVE_UTC and not spam.tzinfo:
                    spam = spam.replace(tzinfo=timezone.utc)
                if (
                    option & OPT_UTC_Z
                    and spam.tzinfo
                    and not spam.tzinfo.utcoffset(spam)
                ):
                    return f"{spam.isoformat()[:-6]}Z"
        return spam.isoformat()  # type: ignore[attr-defined]
    if isinstance(spam, Enum):
        return spam.value
    if type(spam) is UUID:  # pylint: disable=unidiomatic-typecheck
        return str(spam)
    if default:
        return default(spam)

    raise TypeError


def dumps(
    spam: Any,
    /,
    default: None | Callable[[Any], Any] = None,
    option: None | int = None,
) -> bytes:
    """Nobody inspects the spammish repetition."""
    # TODO: only serialize subclasses of str, int, dict, list and Enum
    option = option or 0
    return (
        _dumps(
            spam,
            allow_nan=False,
            default=partial(_default, default, option or 0),
            ensure_ascii=False,
            indent=(option & OPT_INDENT_2) * 2 or None,
            separators=None if option & OPT_INDENT_2 else (",", ":"),
            sort_keys=bool(option & OPT_SORT_KEYS),
        )
        + ("\n" if option & OPT_APPEND_NEWLINE else "")
    ).encode("UTF-8")


def loads(spam: bytes | bytearray | memoryview | str, /) -> Any:
    """Nobody inspects the spammish repetition."""
    if isinstance(spam, memoryview):
        spam = bytes(spam)
    return _loads(spam)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The utilities package with many helpful things used by other modules."""

from __future__ import annotations

import sys

from an_website.utils.static_file_from_traversable import (
    TraversableStaticFileHandler,
)

from .. import DIR as ROOT_DIR
from .utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    # pylint: disable-next=import-outside-toplevel
    from .request_handler import ErrorPage, NotFoundHandler, ZeroDivision

    return ModuleInfo(
        name="Utilities",
        description="Nützliche Werkzeuge für alle möglichen Sachen.",
        handlers=(
            (
                r"/error",
                ZeroDivision if sys.flags.dev_mode else NotFoundHandler,
                {},
            ),
            (r"/([1-5][0-9]{2}).html?", ErrorPage, {}),
            (
                r"/@apm-rum/(.*)",
                TraversableStaticFileHandler,
                {"root": ROOT_DIR / "vendored/apm-rum", "hashes": {}},
            ),
        ),
        hidden=True,
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for commitment."""

from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from datetime import UTC, datetime
from pathlib import Path

import emoji
from time_machine import travel

from an_website.commitment import commitment

from . import (  # noqa: F401  # pylint: disable=unused-import
    DIR,
    FetchCallable,
    app,
    assert_valid_json_response,
    assert_valid_response,
    assert_valid_yaml_response,
    fetch,
)


@contextmanager
def patch_commits_data() -> Iterator[object]:
    """Patch the commits data."""
    commits = commitment.COMMITS
    file = Path(DIR) / "commitment.txt"
    data = commitment.parse_commits_txt(file.read_text())
    if commits is not None:
        assert set(commits) & set(data)
    commitment.COMMITS = data

    try:
        yield None
    finally:
        commitment.COMMITS = commits


def test_parsing() -> None:
    """Test parsing the commitment data."""
    with patch_commits_data():
        assert (data := commitment.COMMITS)

        assert len(list(data)) == 2

        assert data["50821273052022fbc283e310e09168dc65fb3cce"] == (
            datetime(2022, 8, 29, 19, 56, 6, tzinfo=UTC),
            "💬 fix kangaroo comic of today",
        )

        assert data["7335914237808031fa15f32a854ba1e6b1544420"] == (
            datetime(2021, 7, 21, 22, 29, 26, tzinfo=UTC),
            "no_js → no_3rd_party",
        )


async def test_text_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the commitment API."""
    assert commitment.COMMITS, "Please create commits.txt"
    for _ in range(100):  # do it many times to make sure it's not flaky
        response = assert_valid_response(
            await fetch("/api/commitment", headers={"Accept": "*/*"}),
            "text/plain;charset=utf-8",
        )

        assert response.body.decode("UTF-8").endswith("\n")

        response = assert_valid_response(
            await fetch(
                "/api/commitment?require_emoji=sure", headers={"Accept": "*/*"}
            ),
            "text/plain;charset=utf-8",
        )

        assert response.body.decode("UTF-8").endswith("\n")
        assert emoji.emoji_count(response.body.decode("UTF-8"))


async def test_text_api_patched(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the commitment API."""
    with patch_commits_data():
        response = assert_valid_response(
            await fetch("/api/commitment", headers={"Accept": "*/*"}),
            "text/plain;charset=utf-8",
        )

        assert response.body.decode("UTF-8").endswith("\n")
        assert response.body.decode("UTF-8") in {
            "💬 fix kangaroo comic of today\n",
            "no_js → no_3rd_party\n",
        }

        response = assert_valid_response(
            await fetch(
                "/api/commitment?require_emoji=sure", headers={"Accept": "*/*"}
            ),
            "text/plain;charset=utf-8",
        )

        assert response.body.decode("UTF-8").endswith("\n")
        assert (
            response.body.decode("UTF-8") == "💬 fix kangaroo comic of today\n"
        )


async def test_json_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the JSON API."""
    with patch_commits_data():
        for query in (
            "hash=5&require_emoji=sure",
            "hash=50821",
            "hash=50821273052022",
            "hash=50821273052022fbc283e310e09168dc65fb3cce",
        ):
            response = assert_valid_json_response(
                await fetch(
                    f"/api/commitment?{query}",
                    headers={"Accept": "application/json"},
                ),
            )
            assert response["permalink"].endswith(
                "/api/commitment?hash=50821273052022fbc283e310e09168dc65fb3cce"
            )
            del response["permalink"]
            assert response == {
                "commit_message": "💬 fix kangaroo comic of today",
                "hash": "50821273052022fbc283e310e09168dc65fb3cce",
                "date": "2022-08-29T19:56:06Z",
            }


@travel(datetime(2021, 7, 21, 22, 29, 26, tzinfo=UTC), tick=False)
async def test_yaml_api(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the YAML API."""
    with patch_commits_data():
        for query in (
            "hash=7",
            "hash=733591",
            "hash=7335914237808031",
            "hash=7335914237808031fa15f32a854ba1e6b1544420",
        ):
            response = assert_valid_yaml_response(
                await fetch(
                    f"/api/commitment?{query}",
                    headers={"Accept": "application/yaml"},
                ),
            )
            assert response["permalink"].endswith(
                "/api/commitment?hash=7335914237808031fa15f32a854ba1e6b1544420"
            )
            del response["permalink"]
            assert len(response) == 3
            assert response == {
                "commit_message": "no_js → no_3rd_party",
                "date": datetime.now(UTC),
                "hash": "7335914237808031fa15f32a854ba1e6b1544420",
            }


async def test_api_404(fetch: FetchCallable) -> None:  # noqa: F811
    """Test getting non-existent hashes."""
    with patch_commits_data():
        for query in (
            "hash=7&require_emoji=sure",
            "hash=7335914237808031fa15f32a854ba1e6b154442&require_emoji=sure",
            "hash=9168045768942198",
            "hash=9168045768942198e1fbe5e3f00e133b5377a5fa",
            "hash=9168045768942198e1fbe5e3f00e133b5377a5fa6",  # too long
            "hash=x",
            "hash=xxxxxxxxxxxxxxxx",
            "hash=xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "hash=xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",  # too long
        ):
            response = assert_valid_response(
                await fetch(
                    f"/api/commitment?{query}", headers={"Accept": "*/*"}
                ),
                "text/plain;charset=utf-8",
                codes={404},
            )
            assert response.body == b"404 Not Found\n"



# This file is based on the StaticFileHandler from Tornado
# Source: https://github.com/tornadoweb/tornado/blob/b3f2a4bb6fb55f6b1b1e890cdd6332665cfe4a75/tornado/web.py  # noqa: B950  # pylint: disable=line-too-long
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A static file handler for the Traversable abc."""
from __future__ import annotations

import contextlib
import logging
import sys
from collections.abc import Awaitable, Iterable, Mapping, Sequence
from importlib.resources.abc import Traversable
from types import MappingProxyType
from typing import Any, Final, Literal, override
from urllib.parse import urlsplit, urlunsplit

from tornado import httputil, iostream
from tornado.web import GZipContentEncoding, HTTPError
from typed_stream import Stream

from an_website.utils.utils import size_of_file

from .base_request_handler import _RequestHandler
from .static_file_handling import content_type_from_path

type Encoding = Literal["gz", "zst"]

LOGGER: Final = logging.getLogger(__name__)

ENCODINGS: Final[Sequence[tuple[str, Encoding]]] = (
    # better first
    ("zstd", "zst"),
    ("gzip", "gz"),
)
REVERSE_ENCODINGS_MAP: Mapping[Encoding, str] = {
    value: key for key, value in ENCODINGS
}


class TraversableStaticFileHandler(_RequestHandler):
    """A static file handler for the Traversable abc."""

    root: Traversable
    file_hashes: Mapping[str, str] = {}
    headers: Iterable[tuple[str, str]] = ()

    @override
    def compute_etag(self) -> None | str:
        """Return a pre-computed ETag."""
        return self.file_hashes.get(self.request.path)

    async def get(self, path: str, *, head: bool = False) -> None:  # noqa: C901
        # pylint: disable=too-complex, too-many-branches, too-many-statements
        """Handle GET requests for files in the static file directory."""
        if self.request.path.endswith("/"):
            self.replace_path_with_redirect(self.request.path.rstrip("/"))
            return

        if path.startswith("/") or ".." in path.split("/") or "//" in path:
            raise HTTPError(404)

        absolute_path, encoding = self.get_absolute_path_encoded(path)

        if not absolute_path.is_file():
            if self.get_absolute_path(path.lower()).is_file():
                if self.request.path.endswith(path):
                    self.replace_path_with_redirect(
                        self.request.path.removesuffix(path) + path.lower()
                    )
                    return
                LOGGER.error(
                    "Failed to fix casing of %s", self.request.full_url()
                )
            raise HTTPError(404)

        self.set_header("Accept-Ranges", "bytes")

        if encoding:
            self.set_header("Content-Encoding", REVERSE_ENCODINGS_MAP[encoding])
        if content_type := self.get_content_type(
            path, self.get_absolute_path(path)
        ):
            self.set_header("Content-Type", content_type)
        del path

        request_range = None
        range_header = self.request.headers.get("Range")
        if range_header:
            # As per RFC 2616 14.16, if an invalid Range header is specified,
            # the request will be treated as if the header didn't exist.
            # pylint: disable-next=protected-access
            request_range = httputil._parse_request_range(range_header)

        size: int = size_of_file(absolute_path)

        if request_range:
            start, end = request_range
            if start is not None and start < 0:
                start += size
                start = max(start, 0)
            if (
                start is not None
                and (start >= size or (end is not None and start >= end))
            ) or end == 0:  # pylint: disable=use-implicit-booleaness-not-comparison-to-zero # noqa: B950
                # As per RFC 2616 14.35.1, a range is not satisfiable only: if
                # the first requested byte is equal to or greater than the
                # content, or when a suffix with length 0 is specified.
                # https://tools.ietf.org/html/rfc7233#section-2.1
                # A byte-range-spec is invalid if the last-byte-pos value is present
                # and less than the first-byte-pos.
                self.set_status(416)  # Range Not Satisfiable
                self.set_header("Content-Type", "text/plain")
                self.set_header("Content-Range", f"bytes */{size}")
                return
            if end is not None and end > size:
                # Clients sometimes blindly use a large range to limit their
                # download size; cap the endpoint at the actual file size.
                end = size
            # Note: only return HTTP 206 if less than the entire range has been
            # requested. Not only is this semantically correct, but Chrome
            # refuses to play audio if it gets an HTTP 206 in response to
            # ``Range: bytes=0-``.
            if size != (end or size) - (start or 0):
                self.set_status(206)  # Partial Content
                self.set_header(
                    "Content-Range",
                    # pylint: disable-next=protected-access
                    httputil._get_content_range(start, end, size),
                )
        else:
            start = end = None

        content_length = len(range(size)[start:end])
        self.set_header("Content-Length", content_length)

        if head:
            assert self.request.method == "HEAD"
            await self.finish()
            return

        for chunk in self.get_content(absolute_path, start=start, end=end):
            self.write(chunk)
            try:
                await self.flush()
            except iostream.StreamClosedError:
                return
            except httputil.HTTPOutputError:
                LOGGER.exception(
                    "Connection %s; %s",
                    self.request.connection,
                    getattr(self.request.connection, "__dict__", None),
                )
                raise

        with contextlib.suppress(iostream.StreamClosedError):
            await self.finish()

    def get_absolute_path(self, path: str) -> Traversable:
        """Get the absolute path of a file."""
        return self.root / path

    def get_absolute_path_encoded(
        self, path: str
    ) -> tuple[Traversable, Encoding | None]:
        """Get the absolute path and the encoding."""
        for transform in self._transforms:
            if isinstance(transform, GZipContentEncoding):
                # pylint: disable=protected-access
                transform._gzipping = False

        accepted_encodings: frozenset[str] = (
            Stream(self.request.headers.get_list("Accept-Encoding"))
            .flat_map(str.split, ",")
            .map(lambda string: string.split(";")[0])  # ignore quality specs
            .map(str.strip)
            .collect(frozenset)
        )

        absolute_path = self.get_absolute_path(path)
        encoding: Encoding | None = None

        for key, encoding in ENCODINGS:
            if key in accepted_encodings:
                compressed_path = self.get_absolute_path(f"{path}.{encoding}")
                if compressed_path.is_file():
                    absolute_path = compressed_path
                    break
            encoding = None  # pylint: disable=redefined-loop-name

        return absolute_path, encoding

    @classmethod
    def get_content(
        cls,
        abspath: Traversable,
        start: int | None = None,
        end: int | None = None,
    ) -> Iterable[bytes]:
        """Read the content of a file in chunks."""
        with abspath.open("rb") as file:
            if start is not None:
                file.seek(start)
            remaining: int | None = (
                (end - (start or 0)) if end is not None else None
            )

            while True:  # pylint: disable=while-used
                chunk_size = 64 * 1024
                if remaining is not None and remaining < chunk_size:
                    chunk_size = remaining
                chunk = file.read(chunk_size)
                if chunk:
                    if remaining is not None:
                        remaining -= len(chunk)
                    yield chunk
                else:
                    assert not remaining
                    return

    @classmethod
    def get_content_type(
        cls, path: str, absolute_path: Traversable
    ) -> str | None:
        """Get the content-type of a file."""
        return content_type_from_path(path, absolute_path)

    def head(self, path: str) -> Awaitable[None]:
        """Handle HEAD requests for files in the static file directory."""
        return self.get(path, head=True)

    def initialize(
        self,
        root: Traversable,
        hashes: Mapping[str, str] = MappingProxyType({}),
        headers: Iterable[tuple[str, str]] = (),
    ) -> None:
        """Initialize this handler with a root directory and file hashes."""
        self.root = root
        self.file_hashes = hashes
        self.headers = headers
        for name, value in headers:
            self.set_header(name, value)
        if not sys.flags.dev_mode:
            self.set_etag_header()

    def replace_path_with_redirect(
        self, new_path: str, *, status: int = 307
    ) -> None:
        """Redirect to the replaced path."""
        scheme, netloc, _, query, _ = urlsplit(self.request.full_url())
        self.redirect(
            urlunsplit(
                (
                    scheme,
                    netloc,
                    new_path,
                    query,
                    "",
                )
            ),
            status=status,
        )

    @override
    def set_default_headers(self) -> None:
        """Set the default headers for this handler."""
        super().set_default_headers()
        for name, value in self.headers:
            self.set_header(name, value)

        if not sys.flags.dev_mode:
            if "v" in self.request.arguments:
                self.set_header(  # never changes
                    "Cache-Control",
                    f"public, immutable, max-age={86400 * 365 * 10}",
                )
            else:
                self.set_etag_header()

    @override
    def write_error(self, status_code: int, **kwargs: Any) -> None:
        """Write an error response."""
        self.set_header("Content-Type", "text/plain;charset=utf-8")
        self.write(str(status_code))
        self.write(" ")
        self.write(self._reason)
        self.finish("\n")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Parse data into classes."""

from __future__ import annotations

import contextlib
import functools
import inspect
import typing
from collections.abc import Callable, Iterable, Mapping
from inspect import Parameter
from types import UnionType
from typing import Any, Literal, TypeVar, get_origin

from tornado.web import HTTPError, RequestHandler

from .utils import str_to_bool

T = TypeVar("T")


def parse(  # noqa: C901  # pylint: disable=too-many-branches
    type_: type[T],
    data: Iterable[Mapping[str, Any]] | Mapping[str, Any] | Any,
    *,
    strict: bool = False,
) -> T:
    """Parse a data."""
    # pylint: disable=too-many-return-statements, too-complex
    if data is None and type_ is None:
        return None  # type: ignore[unreachable]

    simple_type: type = get_origin(type_) or type_

    if (
        simple_type is type_
        and not isinstance(type_, str)  # type: ignore[redundant-expr]
        and isinstance(data, type_)
    ):
        return data

    if simple_type is Literal:  # type: ignore[comparison-overlap]
        possible = tuple(typing.get_args(type_))
        for pos in possible:
            if pos == data:
                return typing.cast(T, pos)
        if isinstance(data, str):
            data = data.strip()
            for pos in possible:
                if str(pos) == data:
                    return typing.cast(T, pos)
        raise ValueError(f"Unable to parse {data!r} into {type_}")

    if simple_type == UnionType:
        possible = tuple(typing.get_args(type_))
        for pos in possible:
            with contextlib.suppress(ValueError):
                return typing.cast(T, parse(pos, data, strict=strict))
        raise ValueError(f"Unable to parse {data!r} into {type_}")

    if simple_type in {list, "list"} and isinstance(data, list):
        return _parse_list(type_, data, strict=strict)
    if type_ in {bool, "bool"}:
        return typing.cast(T, _parse_bool(data, strict=strict))
    if type_ in {str, "str"}:
        return typing.cast(T, _parse_str(data, strict=strict))
    if type_ in {int, "int"}:
        return typing.cast(T, _parse_int(data, strict=strict))
    if type_ in {float, "float"}:
        return typing.cast(T, _parse_float(data, strict=strict))
    if hasattr(type_, "__init__") and isinstance(data, dict):
        return _parse_class(type_, data, strict=strict)

    raise ValueError(f"Unable to parse {data!r} into {type_}")


def _parse_str(data: Any, *, strict: bool) -> str:
    """Parse data into str."""
    if isinstance(data, str):
        return data
    if strict:
        raise ValueError(f"{data!r} is not str.")
    return str(data)


def _parse_bool(data: Any, *, strict: bool) -> bool:
    """Parse data into bool."""
    if isinstance(data, bool):
        return data
    if strict:
        raise ValueError(f"{data!r} is not bool.")
    if data in {0, 1}:
        return bool(data)
    if isinstance(data, str):
        return str_to_bool(data)
    raise ValueError(f"Cannot parse {data!r} into bool.")


def _parse_int(data: Any, *, strict: bool) -> int:
    """Parse data into int."""
    if isinstance(data, int):
        return data
    if isinstance(data, float) and int(data) == data:
        return int(data)
    if strict:
        raise ValueError(f"{data!r} is not a number.")
    if isinstance(data, str):
        return int(data, base=0)
    if isinstance(data, bool):
        return int(data)
    raise ValueError(f"Cannot parse {data!r} into int.")


def _parse_float(data: Any, *, strict: bool) -> float:
    """Parse data into float."""
    if isinstance(data, float):
        return data
    if isinstance(data, int):
        return float(data)
    if strict:
        raise ValueError(f"{data!r} is not a number.")
    if isinstance(data, str):
        return int(data)
    if isinstance(data, bool):
        return int(data)
    raise ValueError(f"Cannot parse {data!r} into int.")


def _parse_list(
    type_: type[T], data: Iterable[Mapping[str, Any]], *, strict: bool
) -> T:
    """Parse a list of data."""
    args = typing.get_args(type_)
    if len(args) != 1:
        raise ValueError(f"{type_=} should be list[...]")
    return typing.cast(
        T, [parse(args[0], spam, strict=strict) for spam in data]
    )


def _parse_class(type_: type[T], data: Mapping[str, Any], *, strict: bool) -> T:
    """Parse data into a class."""
    signature = inspect.signature(type_.__init__, eval_str=True)
    args: list[Any] = []
    kwargs: dict[str, Any] = {}
    in_positional = False

    def add(_name: str, _value: Any) -> None:
        if in_positional:
            args.append(_value)
        else:
            kwargs[_name] = _value

    first = True
    for arg_name, param in signature.parameters.items():
        if first:
            first = False
            continue
        if param.kind in {Parameter.VAR_KEYWORD, Parameter.KEYWORD_ONLY}:
            in_positional = True
        if arg_name not in data:
            if param.default == Parameter.empty:
                raise ValueError(f"Missing required argument {arg_name!r}")
            add(arg_name, param.default)
            continue
        value = data[arg_name]
        if param.annotation == Parameter.empty:
            if strict:
                raise ValueError(f"Missing type annotation for {arg_name!r}")
            add(arg_name, value)
            continue
        add(arg_name, parse(param.annotation, value, strict=strict))

    return type_(*args, **kwargs)


def parse_args(
    *, type_: Any, name: str = "args", validation_method: str | None = None
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Insert kwarg with name and type into function."""

    def _inner(func: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(func)
        def new_func(self: RequestHandler, *args: Any, **kwargs: Any) -> T:
            arguments: dict[str, str] = {}
            for key, values in self.request.arguments.items():
                if len(values) == 1:
                    arguments[key] = values[0].decode("UTF-8", "replace")
                else:
                    raise HTTPError(  # we don't want to guess
                        400, reason=f"Given multiple values for {key!r}"
                    )
            try:
                _data = parse(type_, arguments, strict=False)
            except ValueError as err:
                raise HTTPError(400, reason=err.args[0]) from err
            if validation_method:
                getattr(_data, validation_method)()
            return func(self, *args, **kwargs, **{name: _data})

        return new_func

    return _inner



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The page that generates art."""

from __future__ import annotations

from collections.abc import Awaitable
from typing import ClassVar, Final

import regex
from redis.asyncio import Redis
from tornado.web import HTTPError

from .. import EVENT_REDIS
from ..utils.base_request_handler import BaseRequestHandler
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    args = r"(?:\d+/)*\d+"
    return ModuleInfo(
        handlers=(
            (r"/LOLWUT", LOLWUT),
            (rf"/LOLWUT/({args})", LOLWUT),
            (r"/api/LOLWUT", LOLWUTAPI),
            (rf"/api/LOLWUT/({args})", LOLWUTAPI),
            (rf"(?i)(?:/api)?/lolwut(?:/{args})?", LOLWUTRedirectHandler),
        ),
        name="LOLWUT",
        description="LOLWUT; präsentiert von Redis",
        path="/LOLWUT",
        keywords=(
            "LOLWUT",
            "Redis",
        ),
    )


def generate_art(redis: Redis[str], args: str | None) -> Awaitable[bytes]:
    """Generate art."""
    return redis.lolwut(*(args.split("/") if args else ()))


class LOLWUT(HTMLRequestHandler):
    """The request handler for the LOLWUT page."""

    async def get(self, args: None | str = None, *, head: bool = False) -> None:
        """Handle GET requests to the LOLWUT page."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)

        art = await generate_art(self.redis, args)

        if head:
            return

        await self.render(
            "ansi2html.html",
            ansi=art,
            powered_by="https://redis.io",
            powered_by_name="Redis",
        )


class LOLWUTAPI(APIRequestHandler):
    """The request handler for the LOLWUT API."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "text/plain",
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
    )

    async def get(self, args: None | str = None) -> None:
        """Handle GET requests to the LOLWUT API."""
        if not EVENT_REDIS.is_set():
            raise HTTPError(503)

        art = await generate_art(self.redis, args)

        if self.content_type == "text/plain":
            return await self.finish(art)

        await self.finish({"LOLWUT": art})


class LOLWUTRedirectHandler(BaseRequestHandler):
    """Redirect to the LOLWUT page."""

    REPL_PATTERN: Final = regex.compile(r"/lolwut(/|\?|$)", regex.IGNORECASE)

    def get(self) -> None:
        """Handle requests to the LOLWUT page."""
        self.redirect(
            LOLWUTRedirectHandler.REPL_PATTERN.sub(
                LOLWUTRedirectHandler.repl_match, self.request.full_url(), 1
            )
        )

    head = get
    post = get

    @staticmethod
    def repl_match(match: regex.Match[str]) -> str:
        """Return the correct replacement for the given match."""
        return f"/LOLWUT{match.group(1)}"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Show a list of all API endpoints."""

from __future__ import annotations

from typing import ClassVar, cast

import orjson as json

from .. import ORJSON_OPTIONS
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo, Permission, name_to_id


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            ("/endpunkte", Endpoints),
            ("/api/endpunkte", EndpointsAPI),
        ),
        name="API-Endpunkte",
        description="Alle API-Endpunkte unserer Webseite",
        path="/endpunkte",
        aliases={
            "/endpoints": "/endpunkte",
            "/api": "/api/endpunkte",
            "/api/endpoints": "/api/endpunkte",
        },
        keywords=("API", "Endpoints", "Endpunkte"),
    )


class Endpoints(HTMLRequestHandler):
    """Endpoint page request handler."""

    async def get(self, *, head: bool = False) -> None:
        """Handle a GET request."""
        if head:
            return
        return await self.render(
            "pages/endpoints.html",
            endpoints=self.get_endpoints(),
            name_to_id=name_to_id,
        )

    def get_endpoints(
        self,
    ) -> list[dict[str, str | list[dict[str, str | list[str]]]]]:
        """Get a list of all API endpoints and return it."""
        endpoints: list[dict[str, str | list[dict[str, str | list[str]]]]] = []
        for module_info in self.settings["MODULE_INFOS"]:
            api_paths: list[dict[str, str | list[str]]] = [
                {
                    "path": handler[0],
                    "methods": methods,
                    "content_types": list(handler[1].POSSIBLE_CONTENT_TYPES),
                }
                for handler in module_info.handlers
                if handler[0].startswith("/api/")
                if issubclass(handler[1], APIRequestHandler)
                if (
                    methods := [
                        method.upper()
                        for method in handler[1].get_allowed_methods()
                        if (
                            perm := getattr(
                                getattr(handler[1], method.lower()),
                                "required_perms",
                                None,
                            )
                        )
                        is None
                        or self.is_authorized(cast(Permission, perm))
                    ]
                )
                if methods != ["OPTIONS"]
            ]
            if len(api_paths) > 0:
                endpoints.append(
                    {
                        "name": module_info.name,
                        "description": module_info.description,
                        "endpoints": api_paths,
                    }
                )
        return endpoints


class EndpointsAPI(APIRequestHandler, Endpoints):
    """Show a list of all API endpoints."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        APIRequestHandler.POSSIBLE_CONTENT_TYPES + ("application/x-ndjson",)
    )

    async def get(self, *, head: bool = False) -> None:
        """Handle a GET request."""
        if self.content_type == "application/x-ndjson":
            await self.finish(
                b"\n".join(
                    json.dumps(endpoint, option=ORJSON_OPTIONS)
                    for endpoint in self.get_endpoints()
                )
            )
        else:
            await self.finish(self.dump(self.get_endpoints()))
